/******************************************************************************
 *  CVS version:
 *     $Id: $
 ******************************************************************************
 *
 *  C code file : symbol.c
 *  Project     : Alan Compiler
 *  Version     : 1.0 alpha
 *  Description : Generic symbol table in C
 *
 *  Comments: (in Greek iso-8859-7)
 *  ---------
 *    .
 *       .
 *      .
 *    
 */


/* ---------------------------------------------------------------------
   ---------------------------- Header files ---------------------------
   --------------------------------------------------------------------- */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include "general.h"
#include "error.h"
#include "types.h"
#include "symbol.h"


/* ---------------------------------------------------------------------
   -------------      --------------
   --------------------------------------------------------------------- */

Scope        * currentScope;           /*                */
unsigned int   quadNext;               /*         */
unsigned int   tempNumber;             /*   temporaries       */

static unsigned int   hashTableSize;   /*    */
static SymbolEntry ** hashTable;       /*          */


/* ---------------------------------------------------------------------
   ------        ------
   --------------------------------------------------------------------- */

void initSymbolTable (unsigned int size)
{
   unsigned int i;

   /*   */

   currentScope = NULL;
   quadNext     = 1;
   tempNumber   = 1;

   /*     */

   hashTableSize = size;
   hashTable = (SymbolEntry **) new(size * sizeof(SymbolEntry *));

   for (i = 0; i < size; i++)
      hashTable[i] = NULL;
}

void destroySymbolTable ()
{
   unsigned int i;

   /*     */

   for (i = 0; i < hashTableSize; i++)
      if (hashTable[i] != NULL)
         destroyEntry(hashTable[i]);

   delete(hashTable);
}

void openScope ()
{
   Scope * newScope = (Scope *) new(sizeof(Scope));

   newScope->negOffset = START_NEGATIVE_OFFSET;
   newScope->parent    = currentScope;
   newScope->entries   = NULL;

   if (currentScope == NULL)
      newScope->nesting = 1;
   else
      newScope->nesting = currentScope->nesting + 1;

   currentScope = newScope;
}

void closeScope ()
{
   SymbolEntry * e = currentScope->entries;
   Scope       * t = currentScope;

   while (e != NULL) {
      unsigned int hashValue = (unsigned int) e->id % hashTableSize;
      SymbolEntry * next = e->nextInScope;

      ASSERT(hashTable[hashValue] == e);
      hashTable[hashValue] = e->nextInHash;
      destroyEntry(e);
      e = next;
   }

   currentScope = currentScope->parent;
   delete(t);
}

static void insertEntry (SymbolEntry * e)
{
   unsigned int hashValue = (unsigned int) e->id % hashTableSize;

   e->nextInHash = hashTable[hashValue];
   hashTable[hashValue]  = e;
   if (currentScope != NULL) {
      e->nextInScope = currentScope->entries;
      currentScope->entries = e;
   }
   else
      e->nextInScope = NULL;
}

static SymbolEntry * newEntry (Identifier id, bool err)
{
   SymbolEntry * e;

   /*        */

   if (err) {
      unsigned int hashValue = (unsigned int) id % hashTableSize;
      unsigned int currentNesting =
         currentScope != NULL ? currentScope->nesting : 0;

      for (e = hashTable[hashValue]; e != NULL; e = e->nextInHash) {
         unsigned int eNesting = e->scope != NULL ? e->scope->nesting : 0;

         if (eNesting < currentNesting)
            break;

         ASSERT(eNesting == currentNesting);
         if (e->id == id) {
            error("duplicate identifier: %s", id_name(id));
            return NULL;
         }
      }
   }

   /*   : entryType  u */

   e = (SymbolEntry *) new(sizeof(SymbolEntry));
   e->id = id;
   e->scope = currentScope;
   insertEntry(e);
   return e;
}

SymbolEntry * newVariable (Identifier id, Type type, bool err)
{
   SymbolEntry * e = newEntry(id, err);

   if (e != NULL) {
      e->entryType = ENTRY_VARIABLE;
      e->u.eVariable.type = type;
      currentScope->negOffset -= sizeOfType(type);
      e->u.eVariable.offset = currentScope->negOffset;
   }
   return e;
}

SymbolEntry * newFunction (Identifier id, bool err)
{
   SymbolEntry * e = lookupEntry(id, LOOKUP_CURRENT_SCOPE, false);

   if (e == NULL) {
      e = newEntry(id, false);
      if (e != NULL) {
         e->entryType = ENTRY_FUNCTION;
         e->u.eFunction.isForward = false;
         e->u.eFunction.pardef = PARDEF_DEFINE;
         e->u.eFunction.firstArgument = e->u.eFunction.lastArgument = NULL;
         e->u.eFunction.resultType = NULL;
      }
      return e;
   }
   else if (e->entryType == ENTRY_FUNCTION && e->u.eFunction.isForward) {
      e->u.eFunction.isForward = false;
      e->u.eFunction.pardef = PARDEF_CHECK;
      e->u.eFunction.lastArgument = NULL;
      return e;
   }
   else {
      if (err)
         error("Duplicate identifier: %s", id_name(id));
      return NULL;
   }
}

SymbolEntry * newParameter (Identifier id, Type type, PassMode mode,
                            SymbolEntry * f, bool err)
{
   SymbolEntry * e;

   if (f->entryType != ENTRY_FUNCTION)
      internal("Cannot add a parameter to a non-function");
   switch (f->u.eFunction.pardef) {
      case PARDEF_DEFINE:
         e = newEntry(id, err);
         if (e != NULL) {
            e->entryType = ENTRY_PARAMETER;
            e->u.eParameter.type = type;
            e->u.eParameter.mode = mode;
            e->u.eParameter.next = NULL;
         }
         if (f->u.eFunction.lastArgument == NULL)
            f->u.eFunction.firstArgument = f->u.eFunction.lastArgument = e;
         else {
            f->u.eFunction.lastArgument->u.eParameter.next = e;
            f->u.eFunction.lastArgument = e;
         }
         return e;
      case PARDEF_CHECK:
         e = f->u.eFunction.lastArgument;
         if (e == NULL)
            e = f->u.eFunction.firstArgument;
         else
            e = e->u.eParameter.next;
         if (e == NULL)
            error("More parameters than expected in redeclaration "
                  "of function %s", f->id);
         else if (!equalType(e->u.eParameter.type, type))
            error("Parameter type mismatch in redeclaration "
                  "of function %s", f->id);
         else if (e->u.eParameter.mode != mode)
            error("Parameter passing mode mismatch in redeclaration "
                  "of function %s", f->id);
         else
            insertEntry(e);
         f->u.eFunction.lastArgument = e;
         return e;
      case PARDEF_COMPLETE:
         fatal("Cannot add a parameter to an already defined function");
   }
   return NULL;
}

static unsigned int fixOffset (SymbolEntry * args)
{
   if (args == NULL)
      return 0;
   else {
      unsigned int rest = fixOffset(args->u.eParameter.next);

      args->u.eParameter.offset = START_POSITIVE_OFFSET + rest;
      if (args->u.eParameter.mode == PASS_BY_REFERENCE)
         return rest + 2;
      else
         return rest + sizeOfType(args->u.eParameter.type);
   }
}

void forwardFunction (SymbolEntry * f)
{
   if (f->entryType != ENTRY_FUNCTION)
      internal("Cannot make a non-function forward");
   f->u.eFunction.isForward = true;
}

void endFunctionHeader (SymbolEntry * f, Type type)
{
   if (f->entryType != ENTRY_FUNCTION)
      internal("Cannot end parameters in a non-function");
   switch (f->u.eFunction.pardef) {
      case PARDEF_COMPLETE:
         internal("Cannot end parameters in an already defined function");
         break;
      case PARDEF_DEFINE:
         fixOffset(f->u.eFunction.firstArgument);
         f->u.eFunction.resultType = type;
         break;
      case PARDEF_CHECK:
         if ((f->u.eFunction.lastArgument != NULL &&
              f->u.eFunction.lastArgument->u.eParameter.next != NULL) ||
             (f->u.eFunction.lastArgument == NULL &&
              f->u.eFunction.firstArgument != NULL))
            error("Fewer parameters than expected in redeclaration "
                  "of function %s", f->id);
         if (!equalType(f->u.eFunction.resultType, type))
            error("Result type mismatch in redeclaration of function %s",
                  f->id);
         break;
   }
   f->u.eFunction.pardef = PARDEF_COMPLETE;
}

SymbolEntry * newTemporary (Type type)
{
   char buffer[10];
   SymbolEntry * e;

   sprintf(buffer, "$%d", tempNumber);
   e = newEntry(id_make(buffer), false);

   if (e != NULL) {
      e->entryType = ENTRY_TEMPORARY;
      e->u.eVariable.type = type;
      currentScope->negOffset -= sizeOfType(type);
      e->u.eTemporary.offset = currentScope->negOffset;
      e->u.eTemporary.number = tempNumber++;
   }
   return e;
}

void destroyEntry (SymbolEntry * e)
{
   SymbolEntry * args;

   switch (e->entryType) {
      case ENTRY_VARIABLE:
         /* nothing */
         break;
      case ENTRY_FUNCTION:
         args = e->u.eFunction.firstArgument;
         while (args != NULL) {
            SymbolEntry * p = args;

            args = args->u.eParameter.next;
            delete(p);
         }
         break;
      case ENTRY_PARAMETER:
         /*        */
         return;
      case ENTRY_TEMPORARY:
         break;
   }
   delete(e);
}

SymbolEntry * lookupEntry (Identifier id, LookupType type, bool err)
{
   unsigned int currentNesting =
      currentScope != NULL ? currentScope->nesting : 0;
   unsigned int hashValue = (unsigned int) id % hashTableSize;
   SymbolEntry * e;

   switch (type) {
      case LOOKUP_CURRENT_SCOPE:
         for (e = hashTable[hashValue]; e != NULL; e = e->nextInHash) {
            unsigned int eNesting = e->scope != NULL ? e->scope->nesting : 0;

            if (eNesting < currentNesting)
               break;

            ASSERT(eNesting == currentNesting);
            if (e->id == id)
               return e;
         }
         break;
      case LOOKUP_ALL_SCOPES:
         for (e = hashTable[hashValue]; e != NULL; e = e->nextInHash)
            if (e->id == id)
               return e;
         break;
   }

   /* ,    */

   if (err)
      error("unknown identifier: %s", id_name(id));
   return NULL;
}
